
/*  Hombre Gauge
    robert.leffmann@gmail.com


    Usage:  Gauge Y "Formatting string"

    Formatting string identifiers:

    %fc %ff %ft   Amount of free chip, fast and total memory
    %Fc %Ff %Ft   printed in abbreviated form

    %vVolume:     Amount of free space on specified volume
    %VVolume:     printed in abbreviated form

    %0 %1 %2 %3   Track position of read head on diskdrive N

    Identifiers from the standard C library strftime() are also recognized.
    See f.ex http://www.cppreference.com/wiki/chrono/c/strftime/

    Setting a Y-position of -1 places the gauge at the bottom of the screen, and
    a position of 0 extends the window height to cover the screen title bar.

    To quit the application, click the gauge with both mousebuttons.


    Thanks to Toni Wilen and Thomas Rapp, and to Olaf Barthel for TrackDisplay.
*/


#include <clib/exec_protos.h>
#include <clib/dos_protos.h>
#include <clib/intuition_protos.h>
#include <clib/graphics_protos.h>
#include <clib/alib_protos.h>

#include <exec/exec.h>
#include <exec/execbase.h>
#include <exec/memory.h>
#include <devices/trackdisk.h>
#include <dos/dos.h>
#include <dos/dosextens.h>
#include <intuition/intuition.h>
#include <graphics/gfxbase.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>


extern struct ExecBase* SysBase;
struct IntuitionBase* IntuitionBase;
struct GfxBase* GfxBase;


typedef struct
{
    struct IOExtTD* request;
    struct MsgPort* port;
    char avail;
} DriveData;

typedef struct
{
  char name[30];
  char value[12];
  unsigned int last;
  char format;
} VolumeData;


static struct
{
  struct Window* window;
  struct RastPort* rport;

  short fh, width, ypos, screenw, screenh, nvolumes, trkoffs;
  unsigned int freechip, freefast;

  struct InfoData* info;

  char formatstring[200];
  char gaugetext[200];

  DriveData drive[4];
  VolumeData volume[10];

} app;


static void ParseCmdLine(const char**);
static void UpdateVolumes(void);
static void FormatText(char*);
static void DrawGauge(void);
static void CloseApp(short);
static void OpenApp(void);
static short DriveStatus(short);
static short PrintSize(char*, unsigned int);




main(int nargs, const char** args)
{
  ParseCmdLine(args);
  OpenApp();

  /*  Main application loop  */

  while(1)
  {
    static short count = 0;
    struct Message* msg;
    short mouse = 0;

    /*  Process all messages  */

    while((msg = GetMsg(app.window->UserPort)) != NULL)
    {
      unsigned int class = ((struct IntuiMessage*)msg)->Class;
      unsigned short code = ((struct IntuiMessage*)msg)->Code;

      ReplyMsg(msg);

      if(class == IDCMP_MOUSEBUTTONS)
      {
        if(code == SELECTDOWN)
          mouse |= 1;
        else if(code == MENUDOWN)
          mouse |= 2;
      }
    };


    /*  Exit if left and right mouse buttons were pressed simultaneously  */

    if(mouse == 3)
      CloseApp(RETURN_OK);

    if(count-- == 0)
    {
      /*  Update memory and volume values once per second only  */

      app.freechip = AvailMem(MEMF_CHIP);
      app.freefast = AvailMem(MEMF_FAST);

      UpdateVolumes();
      count = 10;
    }

    DrawGauge();
    Delay(TICKS_PER_SECOND/10);
  }

}




static void ParseCmdLine(const char** args)
{
  const char* src = args[2];
  char* dst = app.formatstring;
  short vol = 0;
  char c;

  app.ypos = strtol(args[1], NULL, 10);

  do
  {
    c = *(dst++) = *(src++);

    if(c == '%')
    {
      c = src[0];

      if(c == 'v'  ||  c == 'V')
      {
        /*  Copy volume names and translate identifiers into %V0 %V1 etc.  */

        short namelen;

        app.volume[vol].format = dst[0] = c;
        dst[1] = '0'+vol;

        namelen = strchr(src, ':')-src;
        memcpy(app.volume[vol].name, src+1, namelen);

        src += 1+namelen;
        dst += 2;
        ++vol;
      }
    }

  } while(c != 0);

  app.nvolumes = vol;
}




static short PrintSize(char* text, unsigned int n)
{
  char* start = text;
  char size = 0;
  short frac = 0;

  /*  Print integer in abbreviated form  */

  if(n >= 1000000)
  {
    size = 'M';
    n /= 100000;
    frac = n%10;
    n /= 10;
  }
  else if(n >= 1000)
  {
    size = 'K';
    n /= 100;
    frac = n%10;
    n /= 10;
  }

  text += sprintf(text, "%u", n);

  /*  Print a decimal digit for integers less than 10  */

  if(n < 10  &&  frac > 0)
  {
    *(text++) = '.';
    *(text++) = '0'+frac;
  }

  if(size != 0)
    *(text++) = size;

  *text = 0;

  return text-start;
}




static short DriveStatus(short n)
{
  DriveData* d = &app.drive[n];

  /*  Return 1 if disk drive N has a disk inserted, or 0 if not  */

  if(d->avail != 0)
  {
    struct IOStdReq* r = &d->request->iotd_Req;
    short err;

    r->io_Command = TD_CHANGESTATE;
    r->io_Flags = IOF_QUICK;
    err = DoIO((struct IORequest*)d->request);

    if(err == 0  &&  r->io_Actual == 0)
      return 1;
  }

  return 0;
}




static void UpdateVolumes()
{
  struct Process* p = (struct Process*)FindTask(NULL);
  void* temp = p->pr_WindowPtr;
  short n;

  p->pr_WindowPtr = -1;

  for(n = app.nvolumes; n--;)
  {
    VolumeData* v = &app.volume[n];
    BPTR lock = Lock(v->name, ACCESS_READ);
    unsigned int amount = 1;

    if(lock != 0)
    {
      /*  Calculate amount of free space if volume is available  */

      struct InfoData* i = app.info;

      if(Info(lock, i) != 0)
        amount = (i->id_NumBlocks - i->id_NumBlocksUsed) * i->id_BytesPerBlock;

      UnLock(lock);
    }

    if(amount != v->last)
    {
      /*  Update value string if there was a change  */

      v->last = amount;

      if(amount == 1)
      {
        v->value[0] = '-';
        v->value[1] = 0;
      }
      else
      {
        if(v->format == 'v')
          sprintf(v->value, "%u", amount);
        else
          PrintSize(v->value, amount);
      }
    }

  }

  p->pr_WindowPtr = temp;
}




static void FormatText(char* output)
{
  time_t thetime;
  char format[200], c;
  char* pointer = app.formatstring;
  char* temp = format;

  do
  {
    c = *(temp++) = *(pointer++);

    if(c == '%')
    {
      --temp;
      --pointer;
      c = pointer[1];

      if(c == 'f'  ||  c == 'F')
      {
        /*  Print amount of free memory  */

        char memformat = c;
        unsigned int amount;

        c = pointer[2];

        if(c == 'c')
          amount = app.freechip;
        else if (c == 'f')
          amount = app.freefast;
        else if (c == 't')
          amount = app.freechip+app.freefast;

        if(memformat == 'f')
          temp += sprintf(temp, "%u", amount);
        else
          temp += PrintSize(temp, amount);

        pointer += 3;
      }


      else if(c == 'v'  ||  c == 'V')
      {
        /*  Print volume space information  */

        char* value = app.volume[pointer[2]-'0'].value;

        strcpy(temp, value);

        temp += strlen(value);
        pointer += 3;
      }


      else if(c >= '0'  &&  c <= '3')
      {
        /*  Print diskdrive track status  */

        temp[0] = temp[1] = '-';

        if(DriveStatus(c-'0') == 1)
        {
          short track = ((short*)app.drive[c-'0'].request->iotd_Req.io_Unit)[app.trkoffs];
          if(track >= 0)
          {
            track /= 2;
            temp[0] = '0' + track/10;
            temp[1] = '0' + track%10;
          }
        }

        temp += 2;
        pointer += 2;
      }


      else
      {
        /*  Copy other identifiers  */

        temp[1] = c;
        temp += 2;
        pointer += 2;
      }
    }

  } while(c != 0);

  /*  Remaining identifiers are handled by strftime()  */

  time(&thetime);
  strftime(output, 200, format, localtime(&thetime));
}




static void DrawGauge()
{
  static short width = 0;
  char newtext[200];

  FormatText(newtext);

  if(strcmp(app.gaugetext, newtext) != 0)
  {
    /*  Update the gauge if the status text has changed  */

    struct IntuiText t;
    short newwidth;

    strcpy(app.gaugetext, newtext);

    t.FrontPen = 1;
    t.BackPen = 0;
    t.DrawMode = JAM2;
    t.LeftEdge = 0;
    t.TopEdge = 0;
    t.ITextFont = NULL;
    t.IText = newtext;
    t.NextText = NULL;

    newwidth = IntuiTextLength(&t);
    t.LeftEdge = (app.screenw-newwidth)/2;

    if(newwidth < width)
    {
      /*  Erase the sides of the old text if the new text is narrower  */

      unsigned short dif = width-newwidth,
                     l = (app.screenw-width)/2,
                     r = l+width-1;

      RectFill(app.rport, l, 0, l+dif/2, app.fh-1);
      RectFill(app.rport, r-dif/2-1, 0, r, app.fh-1);
    }

    width = newwidth;
    PrintIText(app.rport, &t, 0, 0);
  }

}




static void OpenApp()
{
  struct NewWindow w;
  struct Screen s;
  short n;

  app.info = (struct InfoData*)AllocMem(sizeof(struct InfoData), MEMF_ANY);
  if(app.info == NULL)
    CloseApp(RETURN_FAIL);

  for(n = 4; n--;)
  {
    /*  Allocate and open all trackdisk devices  */

    DriveData* d = &app.drive[n];
    d->avail = 0;

    d->port = (struct MsgPort*)CreatePort(NULL, 0);
    if(d->port != NULL)
    {
      d->request = (struct IOExtTD*)CreateExtIO(d->port, sizeof(struct IOExtTD));
      if(d->request != NULL)
      {
        if(OpenDevice("trackdisk.device", n, (struct IORequest*)d->request, TDF_ALLOW_NON_3_5) == 0)
          d->avail = 1;
      }
    }
  }

  for(n = app.nvolumes; n--;)
  {
    app.volume[n].value[0] = '-';
    app.volume[n].value[1] = 0;
    app.volume[n].last = 1;
  }

  IntuitionBase = (struct IntuitionBase*)OldOpenLibrary("intuition.library");
  GfxBase = (struct GfxBase*)OldOpenLibrary("graphics.library");

  GetScreenData(&s, sizeof(struct Screen), WBENCHSCREEN, NULL);
  app.screenw = s.Width;
  app.fh = GfxBase->DefaultFont->tf_YSize;

  w.LeftEdge    = 0;
  w.TopEdge     = (app.ypos == -1) ? s.Height-app.fh : app.ypos;
  w.Width       = app.screenw;
  w.Height      = (app.ypos == 0) ? s.BarHeight : app.fh;
  w.DetailPen   = -1;
  w.BlockPen    = -1;
  w.IDCMPFlags  = IDCMP_MOUSEBUTTONS;
  w.Flags       = WFLG_BORDERLESS | WFLG_RMBTRAP | WFLG_NOCAREREFRESH | WFLG_SMART_REFRESH;
  w.FirstGadget = NULL;
  w.CheckMark   = NULL;
  w.Title       = NULL;
  w.Type        = WBENCHSCREEN;

  if((app.window = OpenWindow(&w)) == NULL)
    CloseApp(RETURN_FAIL);

  app.trkoffs = (SysBase->LibNode.lib_Version >= 36) ? 27 : 37;
  app.rport = app.window->RPort;
  app.gaugetext[0] = 0;

  SetDrMd(app.rport, JAM1);
  SetAPen(app.rport, 0);
}




static void CloseApp(short err)
{
  short n;

  if(app.window != NULL)
    CloseWindow(app.window);

  /*  Close all trackdisk devices  */

  for(n = 4; n--;)
  {
    DriveData* d = &app.drive[n];

    if(d->avail != 0)
      CloseDevice((struct IORequest*)d->request);

    if(d->request != NULL)
      DeleteExtIO((struct IORequest*)d->request);

    if(d->port != NULL)
      DeletePort(d->port);
  }

  if(app.info != NULL)
    FreeMem(app.info, sizeof(struct InfoData));

  exit(err);
}

